from typing import Any

import json
import yaml


from ._config import Config
from urllib.parse import urlparse
from pathlib import Path
import httpx

from schemax import SchemaData, collect_schema_data

from .output import output
from .utils._cacheir import validate_cache_file, save_cache, load_cache


class SchemaParseError(Exception):
    """Raised when a spec cannot be parsed into a schema."""
    pass

class Spec:
    def __init__(self,
                 spec_link: str,
                 func_name: str,
                 skip_if_failed_to_get_spec: bool = False,
                 is_strict: bool = False,
                 force_strict: bool = False,
                 ):
        self.spec_link = spec_link
        self.func_name = func_name
        self.skip_if_failed_to_get_spec = skip_if_failed_to_get_spec
        self.is_strict = is_strict
        self.force_strict = force_strict

    def _download_spec(self) -> httpx.Response | None:
        def handle_exception(exc: Exception, message: str = ""):
            if self.skip_if_failed_to_get_spec:
                output(func_name=self.func_name, e=exc, text=message)
                return None
            else:
                exc.args = (message,) + exc.args[1:] if exc.args else (message,)
                raise exc
        try:
            response = httpx.get(self.spec_link, timeout=Config.GET_SPEC_TIMEOUT)
            response.raise_for_status()
            return response

        except httpx.ConnectTimeout as e:
            return handle_exception(
                e, f"Timeout occurred while trying to connect to the {self.spec_link}.")
        except httpx.ReadTimeout as e:
            return handle_exception(
                e, f"Timeout occurred while trying to read the spec from the {self.spec_link}.")
        except httpx.HTTPStatusError as e:
            return handle_exception(e)
        except httpx.HTTPError as e:
            return handle_exception(
                e, f"An HTTP error occurred while trying to download the {self.spec_link}")
        except Exception as e:
            return handle_exception(
                e, f"An unexpected error occurred while trying to download the {self.spec_link}")

    def _parse_spec(self, response: httpx.Response) -> dict[str, Any]:
        content_type = response.headers.get('Content-Type', '')

        if 'application/json' in content_type:
            raw_spec = json.loads(response.text)
        elif 'text/yaml' in content_type or 'application/x-yaml' in content_type:
            raw_spec = yaml.load(response.text, Loader=yaml.CLoader)
        else:
            # trying to match via file extension
            if self.spec_link.endswith('.json'):
                raw_spec = json.loads(response.text)
            elif self.spec_link.endswith('.yaml') or self.spec_link.endswith('.yml'):
                raw_spec = yaml.load(response.text, Loader=yaml.CLoader)
            else:
                raise ValueError(f"Unsupported content type: {content_type}")
        return raw_spec

    def _get_schema_from_json(self, raw_spec: dict[str, Any]) -> list[SchemaData]:
        try:
            schema_data = collect_schema_data(raw_spec)
        except Exception as e:
            raise SchemaParseError(
                f"Failed to parse {self.spec_link} to schema via schemax.\n"
                f"Probably the spec is broken or has an unsupported format.\n"
                f"Original exception: {e}")
        return schema_data

    def _build_dict_of_schemas(self, schema_data: list[SchemaData]) -> dict[tuple[str, str, str], SchemaData]:
        entity_dict = {}
        if len(schema_data) == 0:
            raise ValueError("Empty list of entities provided.")
        for elem in schema_data:
            if not isinstance(elem, SchemaData):
                raise TypeError(f"Expected SchemaData, got {type(elem)}")
            entity_key = (elem.http_method.upper(), elem.path, elem.status)
            entity_dict[entity_key] = elem
        return entity_dict

    def _get_raw_spec_from_file(self) -> dict[str, Any]:
        path = Path(self.spec_link)
        if not path.exists():
            raise FileNotFoundError(f"Specification file not found: {self.spec_link}")

        with open(path, 'r') as f:
            if path.suffix == '.json':
                raw_spec = json.loads(f.read())
            elif path.suffix in ('.yml', '.yaml'):
                raw_spec = yaml.load(f.read(), Loader=yaml.CLoader)
            else:
                raise ValueError(f"Unsupported file format: {path.suffix}")
        return raw_spec

    def get_prepared_spec_units(self) -> dict[tuple[str, str, str], SchemaData] | None:
        if self.spec_link is None:
            raise ValueError("Spec link cannot be None")

        if urlparse(self.spec_link).scheme in ('http', 'https', 'ftp') and urlparse(self.spec_link).netloc:
            if validate_cache_file(self.spec_link):
                raw_spec = load_cache(self.spec_link)
            else:
                response = self._download_spec()
                if response is None:
                    return None
                raw_spec = self._parse_spec(response)
                save_cache(spec_link=self.spec_link, raw_schema=raw_spec)
            schema_data = self._get_schema_from_json(raw_spec)
            return self._build_dict_of_schemas(schema_data)
        elif Path(self.spec_link).is_absolute():
            raw_spec = self._get_raw_spec_from_file()
            schema_data = self._get_schema_from_json(raw_spec)
            return self._build_dict_of_schemas(schema_data)
        else:
            raise ValueError(f"{self.spec_link} is neither a valid URL nor a valid path")
        


import unittest
from unittest.mock import patch, Mock, call
import sys
from io import StringIO

from vedro_spec_validator.jj_spec_validator.output import output
from vedro_spec_validator.jj_spec_validator._config import Config


class TestOutput(unittest.TestCase):
    def setUp(self):
        self.original_stdout = sys.stdout
        self.stdout_mock = StringIO()
        sys.stdout = self.stdout_mock
        
        self.original_output_function = Config.OUTPUT_FUNCTION
    
    def tearDown(self):
        sys.stdout = self.original_stdout
        
        Config.OUTPUT_FUNCTION = self.original_output_function
    
    def test_output_with_text_and_exception(self):
        Config.OUTPUT_FUNCTION = None
        func_name = "test_func"
        text = "Error message"
        exception = Exception("Test exception")
        
        output(func_name=func_name, text=text, e=exception)
        
        output_content = self.stdout_mock.getvalue()
        self.assertIn(text, output_content)
        self.assertIn(f"Exception: {str(exception)}", output_content)
    
    def test_output_with_exception_only(self):
        Config.OUTPUT_FUNCTION = None
        func_name = "test_func"
        exception = Exception("Test exception")
        
        output(func_name=func_name, e=exception)
        
        output_content = self.stdout_mock.getvalue()
        self.assertIn(f"Exception: {str(exception)}", output_content)
        self.assertNotIn("None", output_content)
    
    def test_output_with_text_only(self):
        Config.OUTPUT_FUNCTION = None
        func_name = "test_func"
        text = "Info message"
        
        output(func_name=func_name, text=text)
        
        output_content = self.stdout_mock.getvalue()
        self.assertIn(text, output_content)
        self.assertNotIn("Exception", output_content)
    
    def test_output_with_custom_function(self):
        mock_output_function = Mock()
        Config.OUTPUT_FUNCTION = mock_output_function
        
        func_name = "test_func"
        text = "Custom message"
        exception = Exception("Custom exception")
        
        output(func_name=func_name, text=text, e=exception)
        
        mock_output_function.assert_called_once_with(func_name, text, exception)
        
        output_content = self.stdout_mock.getvalue()
        self.assertEqual("", output_content)
    
    def test_output_with_empty_inputs(self):
        Config.OUTPUT_FUNCTION = None
        
        output()
        
        output_content = self.stdout_mock.getvalue()
        self.assertIn("None\n", output_content)
    
    def test_output_with_func_name_only(self):
        Config.OUTPUT_FUNCTION = None
        func_name = "test_func"
        
        output(func_name=func_name)
        
        output_content = self.stdout_mock.getvalue()
        self.assertIn("None\n", output_content)
        self.assertNotIn(func_name, output_content)
    
    @patch('builtins.print')
    def test_output_print_called_correctly(self, mock_print):
        Config.OUTPUT_FUNCTION = None
        func_name = "test_func"
        text = "Message"
        exception = Exception("Test exception")
        
        output(func_name=func_name, text=text, e=exception)
        output(func_name=func_name, text=text)
        output(func_name=func_name, e=exception)
        
        self.assertEqual(mock_print.call_count, 3)
        mock_print.assert_has_calls([
            call(f"{text}\nException: {str(exception)}\n"),
            call(f"{text}\n"),
            call(f"\nException: {str(exception)}\n")
        ]) 


import unittest
from unittest.mock import patch, Mock, mock_open, MagicMock, ANY
import httpx
import json
import yaml
from pathlib import Path

from vedro_spec_validator.jj_spec_validator.spec import Spec, SchemaParseError
from vedro_spec_validator.jj_spec_validator._config import Config
from schemax import SchemaData


class TestDownloadSpec(unittest.TestCase):
    def setUp(self):
        self.spec_link = "https://example.com/api/spec.json"
        self.func_name = "test_func"
        
    @patch('httpx.get')
    def test_download_spec_success(self, mock_get):
        mock_response = Mock()
        mock_response.raise_for_status.return_value = None
        mock_get.return_value = mock_response
        
        spec = Spec(self.spec_link, self.func_name)
        result = spec._download_spec()
        
        self.assertEqual(result, mock_response)
        mock_get.assert_called_once_with(self.spec_link, timeout=Config.GET_SPEC_TIMEOUT)
        mock_response.raise_for_status.assert_called_once()
    
    @patch('httpx.get')
    @patch('vedro_spec_validator.jj_spec_validator.spec.output')
    def test_download_spec_connect_timeout_skip(self, mock_output, mock_get):
        error_message = "Connection timeout"
        mock_get.side_effect = httpx.ConnectTimeout(error_message)
        
        spec = Spec(self.spec_link, self.func_name, skip_if_failed_to_get_spec=True)
        result = spec._download_spec()
        
        self.assertIsNone(result)
        mock_output.assert_called_once()
        self.assertEqual(mock_output.call_args.kwargs['func_name'], self.func_name)
        self.assertIsInstance(mock_output.call_args.kwargs['e'], httpx.ConnectTimeout)
        self.assertTrue(f"Timeout occurred while trying to connect to the {self.spec_link}" in mock_output.call_args.kwargs['text'])
    
    @patch('httpx.get')
    def test_download_spec_connect_timeout_raise(self, mock_get):
        error_message = "Connection timeout"
        original_exception = httpx.ConnectTimeout(error_message)
        mock_get.side_effect = original_exception
        
        spec = Spec(self.spec_link, self.func_name, skip_if_failed_to_get_spec=False)
        with self.assertRaises(httpx.ConnectTimeout) as context:
            spec._download_spec()
            
        self.assertTrue(f"Timeout occurred while trying to connect to the {self.spec_link}" in str(context.exception))
    
    @patch('httpx.get')
    @patch('vedro_spec_validator.jj_spec_validator.spec.output')
    def test_download_spec_read_timeout_skip(self, mock_output, mock_get):
        error_message = "Read timeout"
        mock_get.side_effect = httpx.ReadTimeout(error_message)
        
        spec = Spec(self.spec_link, self.func_name, skip_if_failed_to_get_spec=True)
        result = spec._download_spec()
        
        self.assertIsNone(result)
        mock_output.assert_called_once()
        self.assertEqual(mock_output.call_args.kwargs['func_name'], self.func_name)
        self.assertIsInstance(mock_output.call_args.kwargs['e'], httpx.ReadTimeout)
        self.assertTrue(f"Timeout occurred while trying to read the spec from the {self.spec_link}" in mock_output.call_args.kwargs['text'])
    
    @patch('httpx.get')
    def test_download_spec_read_timeout_raise(self, mock_get):
        error_message = "Read timeout"
        original_exception = httpx.ReadTimeout(error_message)
        mock_get.side_effect = original_exception
        
        spec = Spec(self.spec_link, self.func_name, skip_if_failed_to_get_spec=False)
        with self.assertRaises(httpx.ReadTimeout) as context:
            spec._download_spec()
            
        self.assertTrue(f"Timeout occurred while trying to read the spec from the {self.spec_link}" in str(context.exception))
    
    @patch('httpx.get')
    def test_download_spec_http_status_error_raise(self, mock_get):
        error_message = "404 Not Found"
        mock_request = Mock()
        mock_response = Mock()
        original_exception = httpx.HTTPStatusError(error_message, request=mock_request, response=mock_response)
        mock_get.side_effect = original_exception
        
        spec = Spec(self.spec_link, self.func_name, skip_if_failed_to_get_spec=False)
        with self.assertRaises(httpx.HTTPStatusError) as context:
            spec._download_spec()
        
        self.assertEqual(context.exception, original_exception)
    
    @patch('httpx.get')
    @patch('vedro_spec_validator.jj_spec_validator.spec.output')
    def test_download_spec_http_status_error_skip(self, mock_output, mock_get):
        error_message = "404 Not Found"
        mock_request = Mock()
        mock_response = Mock()
        original_exception = httpx.HTTPStatusError(error_message, request=mock_request, response=mock_response)
        mock_get.side_effect = original_exception
        
        spec = Spec(self.spec_link, self.func_name, skip_if_failed_to_get_spec=True)
        result = spec._download_spec()
        
        self.assertIsNone(result)
        mock_output.assert_called_once()
        self.assertEqual(mock_output.call_args.kwargs['func_name'], self.func_name)
        self.assertEqual(mock_output.call_args.kwargs['e'], original_exception)
    
    @patch('httpx.get')
    @patch('vedro_spec_validator.jj_spec_validator.spec.output')
    def test_download_spec_generic_http_error_skip(self, mock_output, mock_get):
        error_message = "HTTP Error"
        original_exception = httpx.HTTPError(error_message)
        mock_get.side_effect = original_exception
        
        spec = Spec(self.spec_link, self.func_name, skip_if_failed_to_get_spec=True)
        result = spec._download_spec()
        
        self.assertIsNone(result)
        mock_output.assert_called_once()
        self.assertEqual(mock_output.call_args.kwargs['func_name'], self.func_name)
        self.assertEqual(mock_output.call_args.kwargs['e'], original_exception)
        self.assertTrue(f"An HTTP error occurred while trying to download the {self.spec_link}" in mock_output.call_args.kwargs['text'])
    
    @patch('httpx.get')
    @patch('vedro_spec_validator.jj_spec_validator.spec.output')
    def test_download_spec_unexpected_error_skip(self, mock_output, mock_get):
        error_message = "Unexpected error"
        original_exception = Exception(error_message)
        mock_get.side_effect = original_exception
        
        spec = Spec(self.spec_link, self.func_name, skip_if_failed_to_get_spec=True)
        result = spec._download_spec()
        
        self.assertIsNone(result)
        mock_output.assert_called_once()
        self.assertEqual(mock_output.call_args.kwargs['func_name'], self.func_name)
        self.assertEqual(mock_output.call_args.kwargs['e'], original_exception)
        self.assertTrue(f"An unexpected error occurred while trying to download the {self.spec_link}" in mock_output.call_args.kwargs['text'])


class TestParseSpec(unittest.TestCase):
    def setUp(self):
        self.spec_link = "https://example.com/api/spec.json"
        self.func_name = "test_func"
        self.spec = Spec(self.spec_link, self.func_name)
    
    @patch('json.loads')
    def test_parse_spec_json_content_type(self, mock_loads):
        expected_result = {"key": "value"}
        mock_loads.return_value = expected_result
        mock_response = Mock()
        mock_response.headers = {"Content-Type": "application/json; charset=utf-8"}
        mock_response.text = '{"key": "value"}'
        
        result = self.spec._parse_spec(mock_response)
        
        self.assertEqual(result, expected_result)
        mock_loads.assert_called_once_with(mock_response.text)
    
    @patch('yaml.load')
    def test_parse_spec_yaml_content_type(self, mock_yaml_load):
        expected_result = {"key": "value"}
        mock_yaml_load.return_value = expected_result
        mock_response = Mock()
        mock_response.headers = {"Content-Type": "text/yaml"}
        mock_response.text = 'key: value'
        
        result = self.spec._parse_spec(mock_response)
        
        self.assertEqual(result, expected_result)
        mock_yaml_load.assert_called_once_with(mock_response.text, Loader=yaml.CLoader)
    
    @patch('yaml.load')
    def test_parse_spec_application_yaml_content_type(self, mock_yaml_load):
        expected_result = {"key": "value"}
        mock_yaml_load.return_value = expected_result
        mock_response = Mock()
        mock_response.headers = {"Content-Type": "application/x-yaml"}
        mock_response.text = 'key: value'
        
        result = self.spec._parse_spec(mock_response)
        
        self.assertEqual(result, expected_result)
        mock_yaml_load.assert_called_once_with(mock_response.text, Loader=yaml.CLoader)
    
    @patch('json.loads')
    def test_parse_spec_json_file_extension(self, mock_loads):
        expected_result = {"key": "value"}
        mock_loads.return_value = expected_result
        mock_response = Mock()
        mock_response.headers = {"Content-Type": "text/plain"}
        mock_response.text = '{"key": "value"}'
        
        self.spec.spec_link = "https://example.com/api/specification.json"
        
        result = self.spec._parse_spec(mock_response)
        
        self.assertEqual(result, expected_result)
        mock_loads.assert_called_once_with(mock_response.text)
    
    @patch('yaml.load')
    def test_parse_spec_yaml_file_extension(self, mock_yaml_load):
        expected_result = {"key": "value"}
        mock_yaml_load.return_value = expected_result
        mock_response = Mock()
        mock_response.headers = {"Content-Type": "text/plain"}
        mock_response.text = 'key: value'
        
        self.spec.spec_link = "https://example.com/api/specification.yaml"
        
        result = self.spec._parse_spec(mock_response)
        
        self.assertEqual(result, expected_result)
        mock_yaml_load.assert_called_once_with(mock_response.text, Loader=yaml.CLoader)
    
    @patch('yaml.load')
    def test_parse_spec_yml_file_extension(self, mock_yaml_load):
        expected_result = {"key": "value"}
        mock_yaml_load.return_value = expected_result
        mock_response = Mock()
        mock_response.headers = {"Content-Type": "text/plain"}
        mock_response.text = 'key: value'
        
        self.spec.spec_link = "https://example.com/api/specification.yml"
        
        result = self.spec._parse_spec(mock_response)
        
        self.assertEqual(result, expected_result)
        mock_yaml_load.assert_called_once_with(mock_response.text, Loader=yaml.CLoader)
    
    def test_parse_spec_unsupported_content_type(self):
        mock_response = Mock()
        mock_response.headers = {"Content-Type": "text/plain"}
        mock_response.text = 'some text'
        
        self.spec.spec_link = "https://example.com/api/spec.txt"
        
        with self.assertRaises(ValueError):
            self.spec._parse_spec(mock_response)


class TestGetSchemaFromJson(unittest.TestCase):
    def setUp(self):
        self.spec_link = "https://example.com/api/spec.json"
        self.func_name = "test_func"
        self.spec = Spec(self.spec_link, self.func_name)
    
    @patch('vedro_spec_validator.jj_spec_validator.spec.collect_schema_data')
    def test_get_schema_from_json_success(self, mock_collect_schema_data):
        raw_spec = {"paths": {"/test": {"get": {"responses": {"200": {}}}}}}
        expected_schema_data = [MagicMock(spec=SchemaData)]
        mock_collect_schema_data.return_value = expected_schema_data
        
        result = self.spec._get_schema_from_json(raw_spec)
        
        self.assertEqual(result, expected_schema_data)
        mock_collect_schema_data.assert_called_once_with(raw_spec)
    
    @patch('vedro_spec_validator.jj_spec_validator.spec.collect_schema_data')
    def test_get_schema_from_json_error(self, mock_collect_schema_data):
        raw_spec = {"invalid": "spec"}
        error_message = "Failed to parse schema"
        original_exception = Exception(error_message)
        mock_collect_schema_data.side_effect = original_exception
        
        with self.assertRaises(SchemaParseError) as context:
            self.spec._get_schema_from_json(raw_spec)
        
        self.assertTrue(f"Failed to parse {self.spec_link} to schema via schemax" in str(context.exception))
        self.assertTrue("Probably the spec is broken or has an unsupported format" in str(context.exception))
        self.assertTrue(f"Original exception: {original_exception}" in str(context.exception))
        
        mock_collect_schema_data.assert_called_once_with(raw_spec)


class TestBuildDictOfSchemas(unittest.TestCase):
    def setUp(self):
        self.spec_link = "https://example.com/api/spec.json"
        self.func_name = "test_func"
        self.spec = Spec(self.spec_link, self.func_name)
    
    def test_build_dict_of_schemas_success(self):
        mock_schema1 = Mock(spec=SchemaData)
        mock_schema1.http_method = "get"
        mock_schema1.path = "/test"
        mock_schema1.status = "200"
        
        mock_schema2 = Mock(spec=SchemaData)
        mock_schema2.http_method = "post"
        mock_schema2.path = "/test"
        mock_schema2.status = "201"
        
        schema_data = [mock_schema1, mock_schema2]
        
        result = self.spec._build_dict_of_schemas(schema_data)
        
        self.assertEqual(len(result), 2)
        self.assertEqual(result[("GET", "/test", "200")], mock_schema1)
        self.assertEqual(result[("POST", "/test", "201")], mock_schema2)
    
    def test_build_dict_of_schemas_empty_list(self):
        schema_data = []
        
        with self.assertRaises(ValueError) as context:
            self.spec._build_dict_of_schemas(schema_data)
        self.assertEqual(str(context.exception), "Empty list of entities provided.")
    
    def test_build_dict_of_schemas_invalid_type(self):
        schema_data = [{"not": "a SchemaData object"}]
        
        with self.assertRaises(TypeError) as context:
            self.spec._build_dict_of_schemas(schema_data)
        self.assertTrue("Expected SchemaData, got" in str(context.exception))


class TestGetRawSpecFromFile(unittest.TestCase):
    def setUp(self):
        self.func_name = "test_func"
    
    @patch('pathlib.Path.exists')
    @patch('builtins.open', new_callable=mock_open, read_data='{"key": "value"}')
    @patch('json.loads')
    def test_get_raw_spec_from_file_json(self, mock_loads, mock_file, mock_exists):
        spec_link = "/path/to/spec.json"
        expected_result = {"key": "value"}
        mock_exists.return_value = True
        mock_loads.return_value = expected_result
        
        spec = Spec(spec_link, self.func_name)
        
        result = spec._get_raw_spec_from_file()
        
        self.assertEqual(result, expected_result)
        mock_exists.assert_called_once()
        mock_file.assert_called_once_with(ANY, 'r')
        self.assertEqual(mock_file.call_args.args[0], Path(spec_link))
        mock_loads.assert_called_once()
        self.assertEqual(mock_loads.call_args.args[0], '{"key": "value"}')
    
    @patch('pathlib.Path.exists')
    @patch('builtins.open', new_callable=mock_open, read_data='key: value')
    @patch('yaml.load')
    def test_get_raw_spec_from_file_yaml(self, mock_yaml_load, mock_file, mock_exists):
        spec_link = "/path/to/spec.yaml"
        expected_result = {"key": "value"}
        mock_exists.return_value = True
        mock_yaml_load.return_value = expected_result
        
        spec = Spec(spec_link, self.func_name)
        
        result = spec._get_raw_spec_from_file()
        
        self.assertEqual(result, expected_result)
        mock_exists.assert_called_once()
        mock_file.assert_called_once_with(ANY, 'r')
        self.assertEqual(mock_file.call_args.args[0], Path(spec_link))
        mock_yaml_load.assert_called_once()
        self.assertEqual(mock_yaml_load.call_args.args[0], 'key: value')
        self.assertEqual(mock_yaml_load.call_args.kwargs['Loader'], yaml.CLoader)
    
    @patch('pathlib.Path.exists')
    @patch('builtins.open', new_callable=mock_open, read_data='key: value')
    @patch('yaml.load')
    def test_get_raw_spec_from_file_yml(self, mock_yaml_load, mock_file, mock_exists):
        spec_link = "/path/to/spec.yml"
        expected_result = {"key": "value"}
        mock_exists.return_value = True
        mock_yaml_load.return_value = expected_result
        
        spec = Spec(spec_link, self.func_name)
        
        result = spec._get_raw_spec_from_file()
        
        self.assertEqual(result, expected_result)
        mock_exists.assert_called_once()
        mock_file.assert_called_once_with(ANY, 'r')
        self.assertEqual(mock_file.call_args.args[0], Path(spec_link))
        mock_yaml_load.assert_called_once()
        self.assertEqual(mock_yaml_load.call_args.args[0], 'key: value')
        self.assertEqual(mock_yaml_load.call_args.kwargs['Loader'], yaml.CLoader)
    
    @patch('pathlib.Path.exists')
    def test_get_raw_spec_from_file_not_found(self, mock_exists):
        spec_link = "/path/to/spec.json"
        mock_exists.return_value = False
        
        spec = Spec(spec_link, self.func_name)
        
        with self.assertRaises(FileNotFoundError) as context:
            spec._get_raw_spec_from_file()
        self.assertTrue(f"Specification file not found: {spec_link}" in str(context.exception))
        
        mock_exists.assert_called_once()
    
    @patch('pathlib.Path.exists')
    @patch('builtins.open', new_callable=mock_open, read_data='data')
    def test_get_raw_spec_from_file_unsupported_format(self, mock_file, mock_exists):
        spec_link = "/path/to/spec.txt"
        mock_exists.return_value = True
        
        spec = Spec(spec_link, self.func_name)
        
        with self.assertRaises(ValueError) as context:
            spec._get_raw_spec_from_file()
        self.assertTrue(f"Unsupported file format: .txt" in str(context.exception))
        
        mock_exists.assert_called_once()
        mock_file.assert_called_once_with(ANY, 'r')
        self.assertEqual(mock_file.call_args.args[0], Path(spec_link))


class TestGetPreparedSpecUnits(unittest.TestCase):
    def setUp(self):
        self.func_name = "test_func"
    
    def test_spec_link_none(self):
        spec = Spec(None, self.func_name)  # type: ignore
        
        with self.assertRaises(ValueError) as context:
            spec.get_prepared_spec_units()
        self.assertEqual(str(context.exception), "Spec link cannot be None")
    
    @patch('vedro_spec_validator.jj_spec_validator.spec.urlparse')
    @patch('vedro_spec_validator.jj_spec_validator.spec.validate_cache_file')
    @patch('vedro_spec_validator.jj_spec_validator.spec.load_cache')
    @patch('vedro_spec_validator.jj_spec_validator.spec.Spec._get_schema_from_json')
    @patch('vedro_spec_validator.jj_spec_validator.spec.Spec._build_dict_of_schemas')
    def test_get_prepared_spec_units_url_with_cache(self, mock_build_dict, mock_get_schema, 
                                                   mock_load_cache, mock_validate_cache, mock_urlparse):
        spec_link = "https://example.com/api/spec.json"
        url_parts = Mock()
        url_parts.scheme = "https"
        url_parts.netloc = "example.com"
        mock_urlparse.return_value = url_parts
        
        mock_validate_cache.return_value = True
        raw_spec = {"paths": {}}
        mock_load_cache.return_value = raw_spec
        
        mock_schema = Mock(spec=SchemaData)
        mock_schema.http_method = "get"
        mock_schema.path = "/test"
        mock_schema.status = "200"
        
        schema_data = [mock_schema]
        mock_get_schema.return_value = schema_data
        
        expected_result = {("GET", "/test", "200"): mock_schema}
        mock_build_dict.return_value = expected_result
        
        spec = Spec(spec_link, self.func_name)
        
        result = spec.get_prepared_spec_units()
        
        self.assertEqual(result, expected_result)
        mock_urlparse.assert_called_with(spec_link)
        mock_validate_cache.assert_called_once_with(spec_link)
        mock_load_cache.assert_called_once_with(spec_link)
        mock_get_schema.assert_called_once_with(raw_spec)
        mock_build_dict.assert_called_once_with(schema_data)
    
    @patch('vedro_spec_validator.jj_spec_validator.spec.urlparse')
    @patch('vedro_spec_validator.jj_spec_validator.spec.validate_cache_file')
    @patch('vedro_spec_validator.jj_spec_validator.spec.Spec._download_spec')
    @patch('vedro_spec_validator.jj_spec_validator.spec.Spec._parse_spec')
    @patch('vedro_spec_validator.jj_spec_validator.spec.save_cache')
    @patch('vedro_spec_validator.jj_spec_validator.spec.Spec._get_schema_from_json')
    @patch('vedro_spec_validator.jj_spec_validator.spec.Spec._build_dict_of_schemas')
    def test_get_prepared_spec_units_url_without_cache(self, mock_build_dict, mock_get_schema, 
                                                      mock_save_cache, mock_parse_spec, 
                                                      mock_download_spec, mock_validate_cache, mock_urlparse):
        spec_link = "https://example.com/api/spec.json"
        url_parts = Mock()
        url_parts.scheme = "https"
        url_parts.netloc = "example.com"
        mock_urlparse.return_value = url_parts
        
        mock_validate_cache.return_value = False
        
        mock_response = Mock()
        mock_download_spec.return_value = mock_response
        
        raw_spec = {"paths": {}}
        mock_parse_spec.return_value = raw_spec
        
        mock_schema = Mock(spec=SchemaData)
        mock_schema.http_method = "get"
        mock_schema.path = "/test"
        mock_schema.status = "200"
        
        schema_data = [mock_schema]
        mock_get_schema.return_value = schema_data
        
        expected_result = {("GET", "/test", "200"): mock_schema}
        mock_build_dict.return_value = expected_result
        
        spec = Spec(spec_link, self.func_name)
        
        result = spec.get_prepared_spec_units()
        
        self.assertEqual(result, expected_result)
        mock_urlparse.assert_called_with(spec_link)
        mock_validate_cache.assert_called_once_with(spec_link)
        mock_download_spec.assert_called_once()
        mock_parse_spec.assert_called_once_with(mock_response)
        mock_save_cache.assert_called_once_with(spec_link=spec_link, raw_schema=raw_spec)
        mock_get_schema.assert_called_once_with(raw_spec)
        mock_build_dict.assert_called_once_with(schema_data)
    
    @patch('vedro_spec_validator.jj_spec_validator.spec.urlparse')
    @patch('vedro_spec_validator.jj_spec_validator.spec.validate_cache_file')
    @patch('vedro_spec_validator.jj_spec_validator.spec.Spec._download_spec')
    def test_get_prepared_spec_units_download_none(self, mock_download_spec, 
                                                 mock_validate_cache, mock_urlparse):
        spec_link = "https://example.com/api/spec.json"
        url_parts = Mock()
        url_parts.scheme = "https"
        url_parts.netloc = "example.com"
        mock_urlparse.return_value = url_parts
        
        mock_validate_cache.return_value = False
        mock_download_spec.return_value = None
        
        spec = Spec(spec_link, self.func_name)
        
        result = spec.get_prepared_spec_units()
        
        self.assertIsNone(result)
        mock_urlparse.assert_called_with(spec_link)
        mock_validate_cache.assert_called_once_with(spec_link)
        mock_download_spec.assert_called_once()
    
    @patch('vedro_spec_validator.jj_spec_validator.spec.urlparse')
    @patch('pathlib.Path.is_absolute')
    @patch('vedro_spec_validator.jj_spec_validator.spec.Spec._get_raw_spec_from_file')
    @patch('vedro_spec_validator.jj_spec_validator.spec.Spec._get_schema_from_json')
    @patch('vedro_spec_validator.jj_spec_validator.spec.Spec._build_dict_of_schemas')
    def test_get_prepared_spec_units_file_path(self, mock_build_dict, mock_get_schema, 
                                              mock_get_raw_spec, mock_is_absolute, mock_urlparse):
        spec_link = "/path/to/spec.json"
        url_parts = Mock()
        url_parts.scheme = ""
        url_parts.netloc = ""
        mock_urlparse.return_value = url_parts
        
        mock_is_absolute.return_value = True
        
        raw_spec = {"paths": {}}
        mock_get_raw_spec.return_value = raw_spec
        
        mock_schema = Mock(spec=SchemaData)
        mock_schema.http_method = "get"
        mock_schema.path = "/test"
        mock_schema.status = "200"
        
        schema_data = [mock_schema]
        mock_get_schema.return_value = schema_data
        
        expected_result = {("GET", "/test", "200"): mock_schema}
        mock_build_dict.return_value = expected_result
        
        spec = Spec(spec_link, self.func_name)
        
        result = spec.get_prepared_spec_units()
        
        self.assertEqual(result, expected_result)
        mock_urlparse.assert_called_with(spec_link)
        mock_is_absolute.assert_called_once()
        mock_get_raw_spec.assert_called_once()
        mock_get_schema.assert_called_once_with(raw_spec)
        mock_build_dict.assert_called_once_with(schema_data)
    
    @patch('vedro_spec_validator.jj_spec_validator.spec.urlparse')
    @patch('pathlib.Path.is_absolute')
    def test_get_prepared_spec_units_invalid_path(self, mock_is_absolute, mock_urlparse):
        spec_link = "invalid/path/spec.json"
        url_parts = Mock()
        url_parts.scheme = ""
        url_parts.netloc = ""
        mock_urlparse.return_value = url_parts
        
        mock_is_absolute.return_value = False
        
        spec = Spec(spec_link, self.func_name)
        
        with self.assertRaises(ValueError) as context:
            spec.get_prepared_spec_units()
        self.assertEqual(str(context.exception), f"{spec_link} is neither a valid URL nor a valid path")
        
        mock_urlparse.assert_called_with(spec_link)
        mock_is_absolute.assert_called_once()
        


from d42.declaration.types import AnySchema, DictSchema, GenericSchema, ListSchema
from d42.utils import is_ellipsis
from niltype import Nil

__all__ = ('get_forced_strict_spec', 'has_ellipsis_in_all_branches')


def get_forced_strict_spec(schema: GenericSchema) -> GenericSchema:
    if isinstance(schema, DictSchema):
        if schema.props.keys is not Nil:
            new_keys = {}
            for k, (v, is_optional) in schema.props.keys.items():
                if not is_ellipsis(k):
                    new_keys[k] = (get_forced_strict_spec(v), is_optional)
            return schema.__class__(schema.props.update(keys=new_keys))
        return schema
    elif isinstance(schema, ListSchema):
        if schema.props.elements is not Nil:
            new_elements = [get_forced_strict_spec(element) for element in schema.props.elements]
            return schema.__class__(schema.props.update(elements=new_elements))
        elif schema.props.type is not Nil:
            new_type = get_forced_strict_spec(schema.props.type)
            return schema.__class__(schema.props.update(type=new_type))
        return schema
    elif isinstance(schema, AnySchema):
        if schema.props.types is not Nil:
            new_types = tuple(get_forced_strict_spec(t) for t in schema.props.types)
            return schema.__class__(schema.props.update(types=new_types))
        return schema
    else:
        return schema

def has_ellipsis_in_all_branches(schema: GenericSchema) -> bool:
    """
    Check if all branches of the schema contain an Ellipsis object.
    Returns True if every branch has at least one Ellipsis, False otherwise.
    
    The function recursively traverses the schema structure and checks if each
    DictSchema in the structure contains at least one Ellipsis. If any DictSchema
    without an Ellipsis is found, the function returns False.
    
    NOTE: If no DictSchema is found in the entire structure, the function returns False.
    """
    def _has_ellipsis_recursive(schema, dict_found=None):
        if dict_found is None:
            dict_found = [False]
            
        if is_ellipsis(schema):
            return True
            
        if isinstance(schema, DictSchema):
            dict_found[0] = True
            
            if schema.props.keys is Nil or not schema.props.keys:
                return False
                
            has_ellipsis_key = any(is_ellipsis(k) for k in schema.props.keys.keys())
            
            if not has_ellipsis_key:
                return False
                
            for k, (v, _) in schema.props.keys.items():
                if not is_ellipsis(k) and not _has_ellipsis_recursive(v, dict_found):
                    return False
                    
            return True
            
        elif isinstance(schema, ListSchema):
            if schema.props.elements is not Nil and schema.props.elements:
                if not schema.props.elements:
                    return False
                    
                for element in schema.props.elements:
                    if not _has_ellipsis_recursive(element, dict_found):
                        return False
                return True
            elif schema.props.type is not Nil:
                return _has_ellipsis_recursive(schema.props.type, dict_found)
            return False
            
        elif isinstance(schema, AnySchema):
            if schema.props.types is Nil or not schema.props.types:
                return False
                
            for t in schema.props.types:
                if not _has_ellipsis_recursive(t, dict_found):
                    return False
                    
            return True
            
        return True
    
    dict_found = [False]
    result = _has_ellipsis_recursive(schema, dict_found)
    
    if not dict_found[0]:
        return False
        
    return result



from ._config import Config


def output(func_name: str = "", text: str = None, e: Exception = None, ) -> None:
    if Config.OUTPUT_FUNCTION is None:
        if text and e:
            print(f"{text}\nException: {str(e)}\n")
        elif e:
            print(f"\nException: {str(e)}\n")
        else:
            print(f"{text}\n")
    else:
        Config.OUTPUT_FUNCTION(func_name, text, e)



import unittest
from unittest.mock import patch, Mock, MagicMock

from json import JSONDecodeError
from d42.validation import ValidationException
from schemax import SchemaData

from vedro_spec_validator.jj_spec_validator.validator import Validator
from vedro_spec_validator.jj_spec_validator.spec import Spec


class TestValidationFailure(unittest.TestCase):
    def setUp(self):
        self.func_name = "test_func"
        self.validator = Validator(
            skip_if_failed_to_get_spec=False,
            is_raise_error=False,
            is_strict=True,
            func_name=self.func_name,
            spec_link="https://example.com/api/spec.json"
        )
    
    @patch('vedro_spec_validator.jj_spec_validator.validator.output')
    def test_validation_failure_not_raise(self, mock_output):
        exception = Exception("Test exception")
        
        self.validator._validation_failure(exception)
        
        mock_output.assert_called_once()
        self.assertEqual(mock_output.call_args.kwargs['func_name'], self.func_name)
        self.assertEqual(mock_output.call_args.kwargs['e'], exception)
        self.assertTrue(f"There are some mismatches in {self.func_name}" in mock_output.call_args.kwargs['text'])

    @patch('vedro_spec_validator.jj_spec_validator.validator.output')
    def test_validation_failure_raise(self, mock_output):
        exception = Exception("Test exception")
        self.validator.is_raise_error = True
        
        with self.assertRaises(ValidationException) as context:
            self.validator._validation_failure(exception)
        
        self.assertTrue(f"There are some mismatches in {self.func_name}" in str(context.exception))
        self.assertTrue(str(exception) in str(context.exception))
        mock_output.assert_called_once()


class TestPrepareValidation(unittest.TestCase):
    def setUp(self):
        self.func_name = "test_func"
        self.spec_link = "https://example.com/api/spec.json"
        self.validator = Validator(
            skip_if_failed_to_get_spec=False,
            is_raise_error=False,
            is_strict=True,
            func_name=self.func_name,
            spec_link=self.spec_link
        )
        
        self.mock_spec = Mock(spec=Spec)
        
        self.mock_handler = Mock()
        self.mock_response = Mock()
        self.mock_matcher = Mock()
        
        self.mock_handler.matcher = self.mock_matcher
        self.mock_handler.response = self.mock_response
        
        self.mocked = Mock()
        self.mocked.handler = self.mock_handler
    
    @patch('vedro_spec_validator.jj_spec_validator.validator.loads')
    @patch('vedro_spec_validator.jj_spec_validator.validator.create_openapi_matcher')
    def test_prepare_validation_json_content(self, mock_create_matcher, mock_loads):
        self.mock_response.content_type = "application/json"
        self.mock_response.get_body.return_value = '{"key": "value"}'
        
        mock_loads.return_value = {"key": "value"}
        
        mock_spec_matcher = Mock()
        mock_spec_matcher.match.side_effect = lambda x: x == ("GET", "/test")
        mock_create_matcher.return_value = mock_spec_matcher
        
        mock_schema_data = Mock(spec=SchemaData)
        
        prepared_spec = {("GET", "/test", "200"): mock_schema_data}
        self.mock_spec.get_prepared_spec_units.return_value = prepared_spec
        
        self.mock_response.status = "200"
        
        spec_unit, mocked_body = self.validator._prepare_validation(self.mocked, self.mock_spec)
        
        self.assertEqual(spec_unit, mock_schema_data)
        self.assertEqual(mocked_body, {"key": "value"})
        mock_loads.assert_called_once_with(self.mock_response.get_body.return_value)
        mock_create_matcher.assert_called_once_with(matcher=self.mock_matcher, prefix=None)
        mock_spec_matcher.match.assert_called_with(("GET", "/test"))
    
    @patch('vedro_spec_validator.jj_spec_validator.validator.create_openapi_matcher')
    def test_prepare_validation_text_content(self, mock_create_matcher):
        self.mock_response.content_type = "text/plain"
        self.mock_response.text = "some text content"
        
        mock_spec_matcher = Mock()
        mock_spec_matcher.match.side_effect = lambda x: x == ("GET", "/test")
        mock_create_matcher.return_value = mock_spec_matcher
        
        mock_schema_data = Mock(spec=SchemaData)
        
        prepared_spec = {("GET", "/test", "200"): mock_schema_data}
        self.mock_spec.get_prepared_spec_units.return_value = prepared_spec
        
        self.mock_response.status = "200"
        
        spec_unit, mocked_body = self.validator._prepare_validation(self.mocked, self.mock_spec)
        
        self.assertEqual(spec_unit, mock_schema_data)
        self.assertEqual(mocked_body, "some text content")
        mock_create_matcher.assert_called_once_with(matcher=self.mock_matcher, prefix=None)
        mock_spec_matcher.match.assert_called_with(("GET", "/test"))
    
    @patch('vedro_spec_validator.jj_spec_validator.validator.loads')
    def test_prepare_validation_invalid_json(self, mock_loads):
        self.mock_response.content_type = "application/json"
        self.mock_response.get_body.return_value = 'invalid json'
        
        mock_loads.side_effect = JSONDecodeError("Invalid JSON", "invalid json", 0)
        
        with self.assertRaises(AssertionError) as context:
            self.validator._prepare_validation(self.mocked, self.mock_spec)
        
        self.assertTrue(f"There is no valid JSON in {self.func_name}" in str(context.exception))
    
    @patch('vedro_spec_validator.jj_spec_validator.validator.create_openapi_matcher')
    def test_prepare_validation_no_valid_matcher(self, mock_create_matcher):
        self.mock_response.content_type = "text/plain"
        self.mock_response.text = "some text content"
        
        mock_create_matcher.return_value = None
        
        with self.assertRaises(AssertionError) as context:
            self.validator._prepare_validation(self.mocked, self.mock_spec)
        
        self.assertTrue(f"There is no valid matcher in {self.func_name}" in str(context.exception))
    
    @patch('vedro_spec_validator.jj_spec_validator.validator.create_openapi_matcher')
    def test_prepare_validation_no_spec_units(self, mock_create_matcher):
        self.mock_response.content_type = "text/plain"
        self.mock_response.text = "some text content"
        
        mock_spec_matcher = Mock()
        mock_create_matcher.return_value = mock_spec_matcher
        
        self.mock_spec.get_prepared_spec_units.return_value = None
        
        spec_unit, mocked_body = self.validator._prepare_validation(self.mocked, self.mock_spec)
        
        self.assertIsNone(spec_unit)
        self.assertIsNone(mocked_body)
    
    @patch('vedro_spec_validator.jj_spec_validator.validator.create_openapi_matcher')
    def test_prepare_validation_multiple_matches(self, mock_create_matcher):
        self.mock_response.content_type = "text/plain"
        self.mock_response.text = "some text content"
        self.mock_response.status = "200"
        
        mock_spec_matcher = Mock()
        mock_spec_matcher.match.side_effect = lambda x: x[0] in ("GET", "POST")
        mock_create_matcher.return_value = mock_spec_matcher
        
        mock_schema_data1 = Mock(spec=SchemaData)
        mock_schema_data2 = Mock(spec=SchemaData)
        
        prepared_spec = {
            ("GET", "/test", "200"): mock_schema_data1,
            ("POST", "/other", "200"): mock_schema_data2
        }
        self.mock_spec.get_prepared_spec_units.return_value = prepared_spec
        
        with self.assertRaises(AssertionError) as context:
            self.validator._prepare_validation(self.mocked, self.mock_spec)
        
        self.assertTrue(f"There is more than 1 matches for mocked API method" in str(context.exception))
        mock_spec_matcher.match.assert_any_call(("GET", "/test"))
        mock_spec_matcher.match.assert_any_call(("POST", "/other"))
    
    @patch('vedro_spec_validator.jj_spec_validator.validator.create_openapi_matcher')
    def test_prepare_validation_no_status_match(self, mock_create_matcher):
        self.mock_response.content_type = "text/plain"
        self.mock_response.text = "some text content"
        self.mock_response.status = "404"
        
        mock_spec_matcher = Mock()
        mock_spec_matcher.match.side_effect = lambda x: x == ("GET", "/test")
        mock_create_matcher.return_value = mock_spec_matcher
        
        mock_schema_data = Mock(spec=SchemaData)
        
        prepared_spec = {("GET", "/test", "200"): mock_schema_data}
        self.mock_spec.get_prepared_spec_units.return_value = prepared_spec
        
        with self.assertRaises(AssertionError) as context:
            self.validator._prepare_validation(self.mocked, self.mock_spec)
        
        error_message = str(context.exception)
        self.assertTrue(f"Mocked API method: {mock_spec_matcher}, with status: 404" in error_message)
        self.assertTrue(f"was not found in the {self.spec_link}" in error_message)
        self.assertTrue("Presented units" in error_message)
        mock_spec_matcher.match.assert_called_with(("GET", "/test"))


class TestValidate(unittest.TestCase):
    def setUp(self):
        self.func_name = "test_func"
        self.spec_link = "https://example.com/api/spec.json"
        self.validator = Validator(
            skip_if_failed_to_get_spec=False,
            is_raise_error=False,
            is_strict=True,
            func_name=self.func_name,
            spec_link=self.spec_link
        )
        
        self.mock_spec = Mock(spec=Spec)
        self.mocked = Mock()
    
    @patch.object(Validator, '_prepare_validation')
    def test_validate_no_decoded_body(self, mock_prepare_validation):
        mock_prepare_validation.return_value = (None, None)
        
        result = self.validator.validate(self.mocked, self.mock_spec)
        
        self.assertIsNone(result)
        mock_prepare_validation.assert_called_once_with(mocked=self.mocked, spec=self.mock_spec)
    
    @patch.object(Validator, '_prepare_validation')
    def test_validate_no_spec_unit(self, mock_prepare_validation):
        mock_prepare_validation.return_value = (None, {"key": "value"})
        
        with self.assertRaises(AssertionError) as context:
            self.validator.validate(self.mocked, self.mock_spec)
        
        error_message = str(context.exception)
        self.assertTrue(f"API method 'None' in the spec_link" in error_message)
        self.assertTrue(f"lacks a response structure for the validation of {self.func_name}" in error_message)
        mock_prepare_validation.assert_called_once_with(mocked=self.mocked, spec=self.mock_spec)
    
    @patch.object(Validator, '_prepare_validation')
    @patch('vedro_spec_validator.jj_spec_validator.validator.has_ellipsis_in_all_branches')
    @patch('vedro_spec_validator.jj_spec_validator.validator.validate_or_fail')
    def test_validate_strict_validation(self, mock_validate_or_fail, mock_has_ellipsis, mock_prepare_validation):
        mock_schema_data = Mock(spec=SchemaData)
        mock_schema_data.response_schema_d42 = {"type": "object"}
        mock_prepare_validation.return_value = (mock_schema_data, {"key": "value"})
        mock_has_ellipsis.return_value = False
        
        result = self.validator.validate(self.mocked, self.mock_spec)
        
        self.assertIsNone(result)
        mock_prepare_validation.assert_called_once_with(mocked=self.mocked, spec=self.mock_spec)
        mock_validate_or_fail.assert_called_once_with(mock_schema_data.response_schema_d42, {"key": "value"})
    
    @patch.object(Validator, '_prepare_validation')
    @patch('vedro_spec_validator.jj_spec_validator.validator.has_ellipsis_in_all_branches')
    @patch('vedro_spec_validator.jj_spec_validator.validator.validate_non_strict')
    def test_validate_non_strict_validation(self, mock_validate_non_strict, mock_has_ellipsis, mock_prepare_validation):
        self.validator.is_strict = False
        
        mock_schema_data = Mock(spec=SchemaData)
        mock_schema_data.response_schema_d42 = {"type": "object"}
        mock_prepare_validation.return_value = (mock_schema_data, {"key": "value"})
        mock_has_ellipsis.return_value = False
        
        result = self.validator.validate(self.mocked, self.mock_spec)
        
        self.assertIsNone(result)
        mock_prepare_validation.assert_called_once_with(mocked=self.mocked, spec=self.mock_spec)
        mock_validate_non_strict.assert_called_once_with(mock_schema_data.response_schema_d42, {"key": "value"})
    
    @patch.object(Validator, '_prepare_validation')
    @patch('vedro_spec_validator.jj_spec_validator.validator.has_ellipsis_in_all_branches')
    @patch('vedro_spec_validator.jj_spec_validator.validator.validate_or_fail')
    @patch('vedro_spec_validator.jj_spec_validator.validator.output')
    def test_validate_with_ellipsis_warning(self, mock_output, mock_validate_or_fail, mock_has_ellipsis, mock_prepare_validation):
        mock_schema_data = Mock(spec=SchemaData)
        mock_schema_data.response_schema_d42 = {"type": "object"}
        mock_prepare_validation.return_value = (mock_schema_data, {"key": "value"})
        mock_has_ellipsis.return_value = True
        
        result = self.validator.validate(self.mocked, self.mock_spec)
        
        self.assertIsNone(result)
        mock_prepare_validation.assert_called_once_with(mocked=self.mocked, spec=self.mock_spec)
        mock_output.assert_called_once()
        self.assertTrue("`force_strict=True` is required" in mock_output.call_args.kwargs['text'])
    
    @patch.object(Validator, '_prepare_validation')
    @patch('vedro_spec_validator.jj_spec_validator.validator.has_ellipsis_in_all_branches')
    @patch('vedro_spec_validator.jj_spec_validator.validator.get_forced_strict_spec')
    @patch('vedro_spec_validator.jj_spec_validator.validator.validate_or_fail')
    def test_validate_force_strict(self, mock_validate_or_fail, mock_get_forced, mock_has_ellipsis, mock_prepare_validation):
        self.validator.force_strict = True
        
        mock_schema_data = Mock(spec=SchemaData)
        mock_schema_data.response_schema_d42 = {"type": "object"}
        mock_prepare_validation.return_value = (mock_schema_data, {"key": "value"})
        mock_has_ellipsis.return_value = True
        mock_get_forced.return_value = {"type": "object", "strict": True}
        
        result = self.validator.validate(self.mocked, self.mock_spec)
        
        self.assertIsNone(result)
        mock_prepare_validation.assert_called_once_with(mocked=self.mocked, spec=self.mock_spec)
        mock_get_forced.assert_called_once_with(mock_schema_data.response_schema_d42)
        mock_validate_or_fail.assert_called_once_with({"type": "object", "strict": True}, {"key": "value"})
    
    @patch.object(Validator, '_prepare_validation')
    @patch('vedro_spec_validator.jj_spec_validator.validator.has_ellipsis_in_all_branches')
    @patch('vedro_spec_validator.jj_spec_validator.validator.get_forced_strict_spec')
    @patch('vedro_spec_validator.jj_spec_validator.validator.validate_or_fail')
    @patch('vedro_spec_validator.jj_spec_validator.validator.output')
    def test_validate_unnecessary_force_strict(self, mock_output, mock_validate_or_fail, mock_get_forced, mock_has_ellipsis, mock_prepare_validation):
        self.validator.force_strict = True
        
        mock_schema_data = Mock(spec=SchemaData)
        mock_schema_data.response_schema_d42 = {"type": "object"}
        mock_prepare_validation.return_value = (mock_schema_data, {"key": "value"})
        mock_has_ellipsis.return_value = False
        mock_get_forced.return_value = {"type": "object", "strict": True}
        
        result = self.validator.validate(self.mocked, self.mock_spec)
        
        self.assertIsNone(result)
        mock_prepare_validation.assert_called_once_with(mocked=self.mocked, spec=self.mock_spec)
        mock_output.assert_called_once()
        self.assertTrue("`force_strict=True` is not required" in mock_output.call_args.kwargs['text'])
    
    @patch.object(Validator, '_prepare_validation')
    @patch('vedro_spec_validator.jj_spec_validator.validator.has_ellipsis_in_all_branches')
    @patch('vedro_spec_validator.jj_spec_validator.validator.validate_or_fail')
    @patch.object(Validator, '_validation_failure')
    def test_validate_validation_exception(self, mock_validation_failure, mock_validate_or_fail, mock_has_ellipsis, mock_prepare_validation):
        mock_schema_data = Mock(spec=SchemaData)
        mock_schema_data.response_schema_d42 = {"type": "object"}
        mock_prepare_validation.return_value = (mock_schema_data, {"key": "value"})
        mock_has_ellipsis.return_value = False
        
        exception = ValidationException("Validation error")
        mock_validate_or_fail.side_effect = exception
        
        result = self.validator.validate(self.mocked, self.mock_spec)
        
        self.assertIsNone(result)
        mock_prepare_validation.assert_called_once_with(mocked=self.mocked, spec=self.mock_spec)
        mock_validation_failure.assert_called_once_with(exception)
        mock_validate_or_fail.assert_called_once_with(mock_schema_data.response_schema_d42, {"key": "value"})
    
    @patch.object(Validator, '_prepare_validation')
    def test_validate_no_response_schema(self, mock_prepare_validation):
        mock_schema_data = Mock(spec=SchemaData)
        mock_schema_data.response_schema_d42 = None
        mock_prepare_validation.return_value = (mock_schema_data, {"key": "value"})
        
        result = self.validator.validate(self.mocked, self.mock_spec)
        
        self.assertIsNone(result)
        mock_prepare_validation.assert_called_once_with(mocked=self.mocked, spec=self.mock_spec)
        self.assertTrue(hasattr(mock_schema_data, 'response_schema_d42'))
    
    @patch.object(Validator, '_prepare_validation')
    @patch.object(Validator, '_validation_failure')
    def test_validate_decoded_body_without_spec_unit(self, mock_validation_failure, mock_prepare_validation):
        mock_body = {"key": "value"}
        mock_prepare_validation.return_value = (None, mock_body)
        
        with self.assertRaises(AssertionError) as context:
            self.validator.validate(self.mocked, self.mock_spec)
        
        error_message = str(context.exception)
        self.assertTrue(f"API method 'None' in the spec_link" in error_message)
        self.assertTrue(f"lacks a response structure for the validation of {self.func_name}" in error_message)
        
        mock_prepare_validation.assert_called_once_with(mocked=self.mocked, spec=self.mock_spec)
        mock_validation_failure.assert_not_called()



import json
from hashlib import md5
from os import makedirs, path, remove
from pickle import dump
from pickle import load as pickle_load
from time import time
from typing import Any, Dict, Tuple


from .._config import Config


CACHE_DIR = Config.MAIN_DIRECTORY + '/_cache_parsed_specs'
CACHE_TTL = 3600  # in second


def _get_cache_filename(url: str) -> str:
    hash_obj = md5(url.encode())
    return path.join(CACHE_DIR, hash_obj.hexdigest() + '.cache')

def validate_cache_file(spec_link: str) -> bool:
    filename = _get_cache_filename(spec_link)
    if not path.isfile(filename):
        return False

    file_age = time() - path.getmtime(filename)

    if file_age > CACHE_TTL:
        remove(filename)
        return False

    return True

def save_cache(spec_link: str, raw_schema: dict[str, Any]) -> None:
    if not spec_link or not spec_link.strip():
        raise ValueError("spec_link must be a non-empty string")
    filename = _get_cache_filename(spec_link)
    makedirs(CACHE_DIR, exist_ok=True)
    with open(filename, 'wb') as f:
        dump(raw_schema, f)

def load_cache(spec_link: str) -> dict[str, Any]:
    filename = _get_cache_filename(spec_link)
    with open(filename, 'rb') as f:
        raw_spec = pickle_load(f)

    return raw_spec



import click
import httpx
import time
from typing import List
from .spec import Spec
from ._config import Config
from .utils._cacheir import save_cache

@click.command()
@click.argument('spec_links', nargs=-1, required=True)
@click.option('--timeout', default=Config.GET_SPEC_TIMEOUT, help='Timeout for downloading specs in seconds')
def cache_specs(spec_links: List[str], timeout: int) -> None:
    """
    Pre-downloads and caches specifications from provided URLs.
    
    SPEC_LINKS: One or more specification URLs to cache
    """
    start_time = time.time()
    click.echo(f"Starting caching of {len(spec_links)} specifications...")
    
    for spec_link in spec_links:
        spec_start_time = time.time()
        try:
            click.echo(f"\nProcessing {spec_link}...")
            spec = Spec(spec_link=spec_link, func_name="cache_specs")
            response = spec._download_spec()
            
            if response is None:
                click.echo(f"❌ Failed to download specification: {spec_link}")
                continue
                
            raw_spec = spec._parse_spec(response)
            save_cache(spec_link=spec_link, raw_schema=raw_spec)
            spec_time = time.time() - spec_start_time
            click.echo(f"✅ Specification successfully cached: {spec_link} (took {spec_time:.2f}s)")
            
        except Exception as e:
            spec_time = time.time() - spec_start_time
            click.echo(f"❌ Error processing {spec_link}: {str(e)} (took {spec_time:.2f}s)")
    
    total_time = time.time() - start_time
    click.echo(f"\nCaching completed! Total time: {total_time:.2f}s")

if __name__ == '__main__':
    cache_specs() 
    


import unittest
from unittest.mock import patch, Mock, MagicMock
import asyncio

from vedro_spec_validator.jj_spec_validator.validate_spec import validate_spec
from vedro_spec_validator.jj_spec_validator._config import Config
from vedro_spec_validator.jj_spec_validator.spec import Spec
from vedro_spec_validator.jj_spec_validator.validator import Validator


class TestValidateSpec(unittest.TestCase):
    def setUp(self):
        self.spec_link = "https://example.com/api/spec.json"
        self.original_is_enabled = Config.IS_ENABLED
        Config.IS_ENABLED = True
    
    def tearDown(self):
        Config.IS_ENABLED = self.original_is_enabled
    
    @patch('vedro_spec_validator.jj_spec_validator.validate_spec.output')
    def test_validate_spec_with_skip_reason(self, mock_output):
        skip_reason = "Test reason"
        
        @validate_spec(
            spec_link=self.spec_link,
            skip_reason=skip_reason
        )
        def test_func():
            return MagicMock()
        
        test_func()
        
        mock_output.assert_called_once()
        self.assertEqual(mock_output.call_args.kwargs['text'], f"test_func is skipped because: {skip_reason}")
    
    @patch('vedro_spec_validator.jj_spec_validator.validate_spec.Spec')
    @patch('vedro_spec_validator.jj_spec_validator.validate_spec.Validator')
    def test_validate_spec_initialization(self, mock_validator_class, mock_spec_class):
        mock_spec = Mock(spec=Spec)
        mock_spec_class.return_value = mock_spec
        
        mock_validator = Mock(spec=Validator)
        mock_validator_class.return_value = mock_validator
        
        @validate_spec(
            spec_link=self.spec_link,
            skip_if_failed_to_get_spec=True,
            is_raise_error=True,
            is_strict=True,
            prefix="/api",
            force_strict=True
        )
        def test_func():
            return MagicMock()
        
        test_func()
        
        mock_spec_class.assert_called_once_with(
            spec_link=self.spec_link,
            func_name="test_func",
            skip_if_failed_to_get_spec=True,
            is_strict=True,
            force_strict=True
        )
        
        mock_validator_class.assert_called_once_with(
            spec_link=self.spec_link,
            func_name="test_func",
            skip_if_failed_to_get_spec=True,
            is_raise_error=True,
            is_strict=True,
            prefix="/api",
            force_strict=True
        )
    
    @patch('vedro_spec_validator.jj_spec_validator.validate_spec.Spec')
    @patch('vedro_spec_validator.jj_spec_validator.validate_spec.Validator')
    def test_validate_spec_uses_config_defaults(self, mock_validator_class, mock_spec_class):
        mock_spec = Mock(spec=Spec)
        mock_spec_class.return_value = mock_spec
        
        mock_validator = Mock(spec=Validator)
        mock_validator_class.return_value = mock_validator
        
        original_skip = Config.SKIP_IF_FAILED_TO_GET_SPEC
        original_raises = Config.IS_RAISES
        original_strict = Config.IS_STRICT
        
        try:
            Config.SKIP_IF_FAILED_TO_GET_SPEC = True
            Config.IS_RAISES = True
            Config.IS_STRICT = True
            
            @validate_spec(spec_link=self.spec_link)
            def test_func():
                return MagicMock()
            
            test_func()
            
            mock_spec_class.assert_called_once_with(
                spec_link=self.spec_link,
                func_name="test_func",
                skip_if_failed_to_get_spec=True,
                is_strict=True,
                force_strict=False
            )
            
            mock_validator_class.assert_called_once_with(
                spec_link=self.spec_link,
                func_name="test_func",
                skip_if_failed_to_get_spec=True,
                is_raise_error=True,
                is_strict=True,
                prefix=None,
                force_strict=False
            )
            
        finally:
            Config.SKIP_IF_FAILED_TO_GET_SPEC = original_skip
            Config.IS_RAISES = original_raises
            Config.IS_STRICT = original_strict
    
    @patch('vedro_spec_validator.jj_spec_validator.validate_spec.Spec')
    @patch('vedro_spec_validator.jj_spec_validator.validate_spec.Validator')
    def test_validate_spec_sync_function(self, mock_validator_class, mock_spec_class):
        mock_spec = Mock(spec=Spec)
        mock_spec_class.return_value = mock_spec
        
        mock_validator = Mock(spec=Validator)
        mock_validator_class.return_value = mock_validator
        
        @validate_spec(spec_link=self.spec_link)
        def test_func(arg1, arg2=None):
            mock = MagicMock()
            mock.arg1 = arg1
            mock.arg2 = arg2
            return mock
        
        result = test_func("test", arg2="value")
        
        self.assertEqual(result.arg1, "test")
        self.assertEqual(result.arg2, "value")
        mock_validator.validate.assert_called_once_with(result, mock_spec)
    
    @patch('vedro_spec_validator.jj_spec_validator.validate_spec.Spec')
    @patch('vedro_spec_validator.jj_spec_validator.validate_spec.Validator')
    @patch('vedro_spec_validator.jj_spec_validator.validate_spec.isinstance')
    def test_validate_spec_relay_response(self, mock_isinstance, mock_validator_class, mock_spec_class):
        mock_spec = Mock(spec=Spec)
        mock_spec_class.return_value = mock_spec
        
        mock_validator = Mock(spec=Validator)
        mock_validator_class.return_value = mock_validator
        
        # Mock isinstance to correctly identify RelayResponse
        def isinstance_side_effect(obj, class_type):
            from vedro_spec_validator.jj_spec_validator.validate_spec import RelayResponse
            if class_type == RelayResponse:
                return True
            # For all other checks, call the real isinstance
            return isinstance(obj, class_type)
            
        mock_isinstance.side_effect = isinstance_side_effect
        
        @validate_spec(spec_link=self.spec_link)
        def test_func():
            mock = MagicMock()
            mock.handler.response = Mock()
            return mock
        
        with patch('builtins.print') as mock_print:
            result = test_func()
            
            mock_print.assert_called_once_with("RelayResponse type is not supported")
            mock_validator.validate.assert_not_called()
    
    @patch('vedro_spec_validator.jj_spec_validator.validate_spec.Spec')
    @patch('vedro_spec_validator.jj_spec_validator.validate_spec.Validator')
    def test_validate_spec_disabled(self, mock_validator_class, mock_spec_class):
        mock_spec = Mock(spec=Spec)
        mock_spec_class.return_value = mock_spec
        
        mock_validator = Mock(spec=Validator)
        mock_validator_class.return_value = mock_validator
        
        Config.IS_ENABLED = False
        
        @validate_spec(spec_link=self.spec_link)
        def test_func():
            return MagicMock()
        
        result = test_func()
        
        mock_validator.validate.assert_not_called()
    
    @patch('vedro_spec_validator.jj_spec_validator.validate_spec.Spec')
    @patch('vedro_spec_validator.jj_spec_validator.validate_spec.Validator')
    def test_validate_spec_async_function(self, mock_validator_class, mock_spec_class):
        mock_spec = Mock(spec=Spec)
        mock_spec_class.return_value = mock_spec
        
        mock_validator = Mock(spec=Validator)
        mock_validator_class.return_value = mock_validator
        
        @validate_spec(spec_link=self.spec_link)
        async def test_func(arg1, arg2=None):
            mock = MagicMock()
            mock.arg1 = arg1
            mock.arg2 = arg2
            return mock
        
        async def run_test():
            result = await test_func("test", arg2="value")
            self.assertEqual(result.arg1, "test")
            self.assertEqual(result.arg2, "value")
            mock_validator.validate.assert_called_once_with(result, mock_spec)
        
        asyncio.run(run_test())


class TestValidateSpecAsyncDetection(unittest.TestCase):
    def test_is_coroutine_function_detection(self):
        async def async_func():
            pass
            
        def sync_func():
            pass
            
        self.assertTrue(asyncio.iscoroutinefunction(async_func))
        self.assertFalse(asyncio.iscoroutinefunction(sync_func)) 
        


from setuptools import find_packages, setup

from vedro_spec_validator.__version__ import __version__


def find_required():
    with open("requirements.txt") as f:
        return f.read().splitlines()


setup(
    name="vedro-spec-validator",
    version=__version__,
    description="Vedro Spec Validator plugin",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Sam Roz",
    author_email="rolez777@gmail.com",
    python_requires=">=3.10",
    url="https://github.com/Maestoz/vedro-spec-validator",
    license="Apache-2.0",
    packages=find_packages(),
    install_requires=find_required(),
    entry_points={
        'console_scripts': [
            'vedro-spec-cache=vedro_spec_validator.jj_spec_validator.cli:cache_specs',
        ],
    },
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ]
)
